from warnings import warn
from typing import Literal, Union, Optional, Tuple, overload
from numbers import Number

from .. import structures as struct

__all__ = [
    'normalize', 'scale', 'standardize', 'unnormalize', 'unscale', 'unstandardize'
]

def _axis_parsing(input : 'struct.Struct', axis : Literal['spectral', 'spatial', 'all']) :
    """ Check if data is constant and if axis is valid. """
    # Check if axis is valid
    axis = axis.lower()
    if axis == 'all' :
        output_type = float
    elif axis == 'spectral' :
        output_type = struct.Profile
    elif axis == 'spatial' :
        output_type = struct.Map
    else :
        raise ValueError("axis must be 'all', 'spectral' or 'spatial', not {axis}")

    # Return parameters type
    return output_type

@overload
def standardize(input : 'struct.Cube', axis : Literal['spatial'],
    training_subset : Optional[Union['struct.Cube', 'struct.Map', 'struct.Profile']] = None,
    alternative_mode : bool = False) -> Tuple['struct.Cube', 'struct.Map', 'struct.Map'] : ...
@overload
def standardize(input : 'struct.Cube', axis : Literal['spectral'],
    training_subset : Optional[Union['struct.Cube', 'struct.Map', 'struct.Profile']] = None,
    alternative_mode : bool = False) -> Tuple['struct.Cube', 'struct.Profile', 'struct.Profile'] : ...
@overload
def standardize(input : 'struct.Cube', axis : Literal['all'] = 'all',
    training_subset : Optional[Union['struct.Cube', 'struct.Map', 'struct.Profile']] = None,
    alternative_mode : bool = False) -> Tuple['struct.Cube', float, float] : ...
@overload
def standardize(input : 'struct.Map', axis : Literal['all'] = 'all',
    training_subset : Optional['struct.Map'] = None,
    alternative_mode : bool = False) -> Tuple['struct.Map', float, float] : ...
@overload
def standardize(input : 'struct.Profile', axis : Literal['all'] = 'all',
    training_subset : Optional['struct.Profile'] = None,
    alternative_mode : bool = False) -> Tuple['struct.Profile', float, float] : ...

def standardize(input : 'struct.Struct', axis : Literal['spectral', 'spatial', 'all'] = 'all',
    training_subset : Optional['struct.Struct'] = None,
    alternative_mode : bool = False) -> Tuple['struct.Struct', ...] :
    """
    If alternative_mode is False, return a modified version of `input` with mean 0 and standart deviation 1.
    If alternative_mode is True, return a modified version of `input` with mean 0 but same standart deviation.
    """
    output_type = _axis_parsing(input, axis)
    
    if training_subset is None :
        training_subset = input.ones_like()

    mu = training_subset.where(input, float('nan')).mean(output_type)
    sigma = training_subset.where(input, float('nan')).std(output_type)
    
    if alternative_mode and isinstance(sigma, Number) :
        if sigma == 0 :
            warn('Input data are constant and cannot be standardized')
            sigma_ = 1
        else :
            sigma_ = sigma
    elif alternative_mode :
        if (sigma == 0).any() :
            warn('Some features of input data are constant and cannot be standardized')
        sigma_ = (sigma == 0).where(1, sigma)

    if alternative_mode :
        out = input - mu
    else :
        out = (input - mu) / sigma_
    
    return out, mu, sigma

@overload
def unstandardize(input : 'struct.Cube', mu : Union['struct.Map', 'struct.Profile', float],
    sigma : Union['struct.Map', 'struct.Profile', float], alternative_mode : bool = False) -> 'struct.Cube' : ...
@overload
def unstandardize(input : 'struct.Map', mu : float,
    sigma : float, alternative_mode : bool = False) -> 'struct.Map' : ...
@overload
def unstandardize(input : 'struct.Profile', mu : float,
    sigma : float, alternative_mode : bool = False) -> 'struct.Profile' : ...

def unstandardize(input : 'struct.Struct', mu : Union['struct.Struct', float],
    sigma : Union['struct.Struct', float], alternative_mode : bool = False) -> 'struct.Struct' :
    """ TODO """
    if alternative_mode :
        return input + mu
    return sigma * input + mu

@overload
def normalize(input : 'struct.Cube', axis : Literal['spatial'],
    training_subset : Optional[Union['struct.Cube', 'struct.Map', 'struct.Profile']] = None,
    alternative_mode : bool = False) -> Tuple['struct.Cube', 'struct.Map', 'struct.Map'] : ...
@overload
def normalize(input : 'struct.Cube', axis : Literal['spectral'],
    training_subset : Optional[Union['struct.Cube', 'struct.Map', 'struct.Profile']] = None,
    alternative_mode : bool = False) -> Tuple['struct.Cube', 'struct.Profile', 'struct.Profile'] : ...
@overload
def normalize(input : 'struct.Cube', axis : Literal['all'] = 'all',
    training_subset : Optional[Union['struct.Cube', 'struct.Map', 'struct.Profile']] = None,
    alternative_mode : bool = False) -> Tuple['struct.Cube', float, float] : ...
@overload
def normalize(input : 'struct.Map', axis : Literal['all'] = 'all',
    training_subset : Optional['struct.Map'] = None,
    alternative_mode : bool = False) -> Tuple['struct.Map', float, float] : ...
@overload
def normalize(input : 'struct.Profile', axis : Literal['all'] = 'all',
    training_subset : Optional['struct.Profile'] = None,
    alternative_mode : bool = False) -> Tuple['struct.Profile', float, float] : ...

def normalize(input : 'struct.Struct', axis : Literal['spectral', 'spatial', 'all'] = 'all',
    training_subset : Optional['struct.Struct'] = None,
    alternative_mode : bool = False) -> Tuple[Union['struct.Struct', float], ...] :
    """
    If alternative_mode is False, return a modified version of `input` with every values between 0 and 1.
    If alternative_mode is True, return a modified version of `input` with every values between -1 and 1.
    """
    output_type = _axis_parsing(input, axis)
    
    if training_subset is None :
        training_subset = input.ones_like()

    a = training_subset.where(input, float('nan')).min(output_type)
    b = training_subset.where(input, float('nan')).max(output_type)

    if isinstance(a, Number) :
        if a == b :
            warn('Input data are constant and cannot be standardized')
            scale_ = 1
        else :
            scale_ = b - a
    else :
        if (a == b).any() :
            warn('Some features of input data are constant and cannot be standardized')
        scale_ = (a == b).where(1, b - a)

    if alternative_mode :
        out = 2 * (input - a) / scale_ - 1 + (a==b).where(1, 0)
    else :
        out = (input - a) / scale_
    
    return out, a, b

@overload
def unnormalize(input : 'struct.Cube', a : Union['struct.Map', 'struct.Profile', float],
    b : Union['struct.Map', 'struct.Profile', float]) -> 'struct.Cube': ...
@overload
def unnormalize(input : 'struct.Map', a : float, b : float) -> 'struct.Map' : ...
@overload
def unnormalize(input : 'struct.Profile', a : float, b : float) -> 'struct.Profile' : ...

def unnormalize(input : 'struct.Struct', a : Union['struct.Map', 'struct.Profile', float],
    b : Union['struct.Struct', float], alternative_mode : bool = False) -> 'struct.Struct' :
    """ TODO """
    return (b-a) * input + a

@overload
def scale(input : 'struct.Cube', axis : Literal['spatial'],
    training_subset : Optional[Union['struct.Cube', 'struct.Map', 'struct.Profile']] = None,
    alternative_mode : bool = False) -> Tuple['struct.Cube', 'struct.Map'] : ...
@overload
def scale(input : 'struct.Cube', axis : Literal['spectral'],
    training_subset : Optional[Union['struct.Cube', 'struct.Map', 'struct.Profile']] = None,
    alternative_mode : bool = False) -> Tuple['struct.Cube', 'struct.Profile'] : ...
@overload
def scale(input : 'struct.Cube', axis : Literal['all'] = 'all',
    training_subset : Optional[Union['struct.Cube', 'struct.Map', 'struct.Profile']] = None,
    alternative_mode : bool = False) -> Tuple['struct.Cube', float] : ...
@overload
def scale(input : 'struct.Map', axis : Literal['all'] = 'all',
    training_subset : Optional['struct.Map'] = None,
    alternative_mode : bool = False) -> Tuple['struct.Map', float] : ...
@overload
def scale(input : 'struct.Profile', axis : Literal['all'] = 'all',
    training_subset : Optional['struct.Map'] = None,
    alternative_mode : bool = False) -> Tuple['struct.Profile', float] : ...

def scale(input : 'struct.Struct', axis : Literal['spectral', 'spatial', 'all'] = 'all',
    training_subset : Optional['struct.Struct'] = None,
    alternative_mode : bool = False) -> Tuple[Union['struct.Map', 'struct.Profile', float], ...] :
    """
    If alternative_mode is False, return a modified version of `input` where every value is between -1 and 1
    If alternative_mode is True, raise a NotImplementedError
    """
    output_type = _axis_parsing(input, axis)
    
    if training_subset is None :
        training_subset = input.ones_like()

    s = training_subset.where(input, float('nan')).abs().max(output_type)

    if isinstance(s, Number) :
        if s == 0 :
            warn('Input data are constant and cannot be standardized')
            s_ = 1
        else :
            s_ = s
    else :
        if (s == 0).any() :
            warn('Some features of input data are constant and cannot be standardized')
        s_ = (s == 0).where(1, s)

    if alternative_mode :
        raise NotImplementedError()
    else :
        pass
    out = input / s

    return out, s

@overload
def unscale(input : 'struct.Cube', s : Union['struct.Map', 'struct.Profile', float]) -> 'struct.Cube': ...
@overload
def unscale(input : 'struct.Map', s : float) -> 'struct.Map' : ...
@overload
def unscale(input : 'struct.Profile', s : float) -> 'struct.Profile' : ...

def unscale(input : 'struct.Struct', s : Union['struct.Struct', float], alternative_mode : bool = False) :
    """ TODO """
    if alternative_mode :
        raise NotImplementedError()
    else :
        return s * input


from abc import ABC, abstractmethod
from warnings import warn

import numpy as np
from astropy.io import fits

from .core import header as hdr

__all__ = ["Cube", "Map", "Profile", "Struct"]


class Struct(ABC):
    """Radio astronomy multidimensional data"""

    def __init__(self, data: np.ndarray, header: fits.Header):
        """
        Initializer.

        Parameters
        ----------
        data : np.ndarray
            Data array.
        header : fits.Header
            FITS header.
        """
        if data.dtype is np.dtype("bool"):
            data = data.astype("float")
        elif data.dtype is np.dtype("int"):
            data = data.astype("float")
        if np.isinf(data).any():
            data[np.isinf(data)] = float("nan")  # Avoid inf in FITS header
        self.data: np.ndarray = data  #: Data array
        self.header: fits.Header = hdr.update_header(data, header)  #: FITS header

    from .core.util import copy, from_fits, from_numpy, ones, zeros

    from_fits = classmethod(from_fits)
    from_numpy = classmethod(from_numpy)
    zeros = classmethod(zeros)
    ones = classmethod(ones)

    @abstractmethod
    def __str__(self) -> str:
        """Returns a printable version of the structure. Can be called with str(self)"""
        pass

    @property
    def shape(self: "Struct") -> tuple[int]:
        """Shape of the data"""
        return self.data.shape

    @property
    def size(self: "Struct") -> int:
        """Number of scalars in the cube."""
        return self.data.size

    # Unary float operators

    # Getitem operator
    # Comparison operators
    # Binary boolean operators
    # Unary boolean operators
    # Binary float operators
    from .core._op import (
        __abs__,
        __add__,
        __and__,
        __ceil__,
        __eq__,
        __floor__,
        __floordiv__,
        __ge__,
        __getitem__,
        __gt__,
        __invert__,
        __le__,
        __lt__,
        __mod__,
        __mul__,
        __ne__,
        __neg__,
        __or__,
        __pow__,
        __radd__,
        __rand__,
        __rmul__,
        __ror__,
        __round__,
        __rpow__,
        __rsub__,
        __rtruediv__,
        __rxor__,
        __sub__,
        __truediv__,
        __xor__,
    )
    from .core.math import (
        abs,
        arccos,
        arccosh,
        arcsin,
        arcsinh,
        arctan,
        arctanh,
        cbrt,
        cos,
        cosh,
        exp,
        log,
        sin,
        sinh,
        sqrt,
        tan,
        tanh,
    )
    from .core.util import (
        apply_element_wise,
        astype,
        clip,
        is_logical,
        isfinite,
        isnan,
        ones_like,
        to_numpy,
        where,
        zeros_like,
    )
    from .filtering.morphology import (
        closing,
        dilation,
        erosion,
        gradient,
        laplacian,
        opening,
    )

    # Others
    from .io.io import _save_fits as save_fits
    from .io.io import (
        plot_hist,
        plot_hist2d,
        save_hist,
        save_hist2d,
        show_hist,
        show_hist2d,
    )
    from .learning.preprocessing import (
        normalize,
        scale,
        standardize,
        unnormalize,
        unscale,
        unstandardize,
    )
    from .models.distribution import kde
    from .models.noise import additive_noise, multiplicative_noise
    from .reduction.stats import (
        all,
        any,
        argmax,
        argmin,
        max,
        mean,
        median,
        min,
        moment,
        percentile,
        ptp,
        quantile,
        rms,
        std,
        sum,
        var,
    )


class Cube(Struct):
    """Radio astronomy data cube"""

    def __init__(self, data: np.ndarray, header: fits.Header):
        # Check data and header number of axis
        if data.ndim != 3:
            raise ValueError(f"data must have 3 dimensions, not {data.ndim}")
        if header["NAXIS"] != 3:
            raise ValueError(f"header must have 3 axes, not {header['NAXIS']}")
        # Check axes compatibility
        dims = (header["NAXIS3"], header["NAXIS2"], header["NAXIS1"])
        if data.shape == (header["NAXIS3"], header["NAXIS2"], header["NAXIS1"]):
            pass
        elif data.shape == (header["NAXIS3"], header["NAXIS1"], header["NAXIS2"]):
            warn(f"Axis of cube data swapped to match shape {dims}")
            data = np.moveaxis(data, (0, 2, 1), (0, 1, 2))
        elif data.shape == (header["NAXIS2"], header["NAXIS1"], header["NAXIS3"]):
            warn(f"Axis of cube data swapped to match shape {dims}")
            data = np.moveaxis(data, (2, 0, 1), (0, 1, 2))
        elif data.shape == (header["NAXIS1"], header["NAXIS2"], header["NAXIS3"]):
            warn(f"Axis of cube data swapped to match shape {dims}")
            data = np.moveaxis(data, (2, 1, 0), (0, 1, 2))
        else:
            raise ValueError(
                f"Shape of data {data.shape} cannot match {dims}, even by swapping axes"
            )
        super().__init__(data, header)

    from .core.util import cube_from_maps as from_maps
    from .core.util import cube_from_profiles as from_profiles

    from_maps = staticmethod(from_maps)
    from_profiles = staticmethod(from_profiles)

    @property
    def nx(self: "Struct") -> int:
        """Length of cube x axis"""
        return self.data.shape[2]

    @property
    def ny(self: "Struct") -> int:
        """Length of cube y axis"""
        return self.data.shape[1]

    @property
    def nz(self: "Struct") -> int:
        """Length of cube z axis"""
        return self.data.shape[0]

    def __str__(self) -> str:
        """Returns a printable version of the cube. Can be called with str(self)"""
        return f"Cube (nx: {self.nx}, ny: {self.ny}, nz: {self.nz})"

    from .core.util import (
        change_axes_order,
        map_from,
        profile_from,
        x_axis,
        y_axis,
        z_axis,
    )
    from .filtering.filters import filter_channels
    from .filtering.filters import filter_cube as filter
    from .filtering.filters import filter_pixels
    from .io.io import (
        plot_channel,
        plot_pixel,
        save_channel_plot,
        save_pixel_plot,
        show_channel,
        show_pixel,
    )
    from .reduction.astro import (
        integral,
        noise_map,
        reduce_spatial,
        reduce_spectral,
        spectrum,
    )
    from .reduction.getters import get_channel, get_channels, get_pixel, get_pixels


class Map(Struct):
    """Radio astronomy data map"""

    def __init__(self, data: np.ndarray, header: fits.Header):
        # Check data and header number of axis
        if data.ndim != 2:
            raise ValueError(f"data must have 2 dimensions, not {data.ndim}")
        if header["NAXIS"] != 2:
            raise ValueError(f"header must have 2 axes, not {header['NAXIS']}")
        # Check axes compatibility
        dims = (header["NAXIS2"], header["NAXIS1"])
        if data.shape == (header["NAXIS2"], header["NAXIS1"]):
            pass
        elif data.shape == (header["NAXIS1"], header["NAXIS2"]):
            warn(f"Axis of map data swapped to match shape {dims}")
            data = data.T
        else:
            raise ValueError(
                f"Shape of data {data.shape} cannot match {dims}, even by swapping axes"
            )
        super().__init__(data, header)

    @property
    def nx(self: "Struct") -> int:
        """Length of cube x axis"""
        return self.data.shape[1]

    @property
    def ny(self: "Struct") -> int:
        """Length of cube y axis"""
        return self.data.shape[0]

    def __str__(self) -> str:
        """Returns a printable version of the map. Can be called with str(self)"""
        return f"Map (nx: {self.nx}, ny: {self.ny})"

    from .core.util import change_axes_order, x_axis, y_axis
    from .filtering.filters import filter_map as filter
    from .io.io import plot_map as plot
    from .io.io import save_map_plot as save_plot
    from .io.io import show_map as show
    from .reduction.astro import reduce_spatial


class Profile(Struct):
    """Radio astronomy data profile"""

    def __init__(self, data: np.ndarray, header: fits.Header):
        # Check data and header number of axis
        if data.ndim != 1:
            raise ValueError(f"data must have 1 dimension, not {data.ndim}")
        if header["NAXIS"] != 1:
            raise ValueError(f"header must have 1 axis, not {header['NAXIS']}")
        super().__init__(data.flatten(), header)

    @property
    def nz(self: "Struct") -> int:
        """Length of cube z axis"""
        return self.data.shape[0]

    def __str__(self) -> str:
        """Returns a printable version of the profile. Can be called with str(self)"""
        return f"Profile (nz: {self.nz})"

    from .core.util import z_axis
    from .filtering.filters import filter_profile as filter
    from .io.io import plot_profile as plot
    from .io.io import save_profile_plot as save_plot
    from .io.io import show_profile as show
    from .reduction.astro import reduce_spectral



from numbers import Integral, Number
from typing import Callable, Tuple, Union, overload
from warnings import warn

import numpy as np

from .. import structures as struct
from . import header as hdr


@overload
def _binary_op(
    a: "struct.Cube",
    b: Union["struct.Cube", "struct.Map", "struct.Profile", float],
    fun: Callable,
) -> "struct.Cube":
    ...


@overload
def _binary_op(
    a: "struct.Map", b: Union["struct.Map", float], fun: Callable
) -> "struct.Map":
    ...


@overload
def _binary_op(
    a: "struct.Map", b: Union["struct.Cube", "struct.Profile"], fun: Callable
) -> "struct.Cube":
    ...


@overload
def _binary_op(
    a: "struct.Profile", b: Union["struct.Profile", float], fun: Callable
) -> "struct.Profile":
    ...


@overload
def _binary_op(
    a: "struct.Profile", b: Union["struct.Cube", "struct.Map"], fun: Callable
) -> "struct.Cube":
    ...


def _binary_op(
    a: "struct.Struct", b: Union["struct.Struct", float], fun: Callable
) -> "struct.Struct":
    """
    Convenience function for binary operators.

    Parameters
    ----------
    a : struct.Struct
        First input.
    b : Union[&#39;struct.Struct&#39;, float]
        Second input.
    fun : Callable
        Implementation of operator to be applied on numpy arrays.

    Returns
    -------
    struct.Struct
        Output multidimensional data.
    """
    if isinstance(b, Number):
        new_data = fun(a.data, b)
        return type(a)(new_data, a.header)
    new_header = hdr.merge_headers(a.header, b.header)
    if type(a) == type(b):
        new_data = fun(a.data, b.data)
        return type(a)(new_data, new_header)
    if isinstance(a, struct.Map) and isinstance(b, struct.Profile):
        new_data = fun(np.expand_dims(a.data, 0), np.expand_dims(b.data, [1, 2]))
    elif isinstance(a, struct.Profile) and isinstance(b, struct.Map):
        new_data = fun(np.expand_dims(a.data, [1, 2]), np.expand_dims(b.data, 0))
    elif isinstance(a, struct.Cube) and isinstance(b, struct.Profile):
        new_data = fun(a.data, np.expand_dims(b.data, [1, 2]))
    elif isinstance(a, struct.Profile) and isinstance(b, struct.Cube):
        new_data = fun(np.expand_dims(a.data, [1, 2]), b.data)
    else:
        new_data = fun(a.data, b.data)
    return struct.Cube(new_data, new_header)


# Unary float operators


@overload
def __neg__(self: "struct.Cube") -> "struct.Cube":
    ...


@overload
def __neg__(self: "struct.Map") -> "struct.Map":
    ...


@overload
def __neg__(self: "struct.Profile") -> "struct.Profile":
    ...


def __neg__(self: "struct.Struct") -> "struct.Struct":
    """Returns -self"""
    return type(self)(-self.data, self.header)


@overload
def __abs__(self: "struct.Cube") -> "struct.Cube":
    ...


@overload
def __abs__(self: "struct.Map") -> "struct.Map":
    ...


@overload
def __abs__(self: "struct.Profile") -> "struct.Profile":
    ...


def __abs__(self: "struct.Struct"):
    """Returns abs(self)"""
    return type(self)(-np.abs(self.data), self.header)


@overload
def __round__(self: "struct.Cube") -> "struct.Cube":
    ...


@overload
def __round__(self: "struct.Map") -> "struct.Map":
    ...


@overload
def __round__(self: "struct.Profile") -> "struct.Profile":
    ...


def __round__(self: "struct.Struct"):
    """Returns round(self)"""
    return type(self)(self.data.round(), self.header)


@overload
def __floor__(self: "struct.Cube") -> "struct.Cube":
    ...


@overload
def __floor__(self: "struct.Map") -> "struct.Map":
    ...


@overload
def __floor__(self: "struct.Profile") -> "struct.Profile":
    ...


def __floor__(self: "struct.Struct"):
    """Returns floor(self)"""
    return type(self)(np.floor(self.data), self.header)


@overload
def __ceil__(self: "struct.Cube") -> "struct.Cube":
    ...


@overload
def __ceil__(self: "struct.Map") -> "struct.Map":
    ...


@overload
def __ceil__(self: "struct.Profile") -> "struct.Profile":
    ...


def __ceil__(self: "struct.Struct"):
    """Returns ceil(self)"""
    return type(self)(np.ceil(self.data), self.header)


# Binary float operators


@overload
def __add__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __add__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __add__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __add__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __add__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __add__(
    self: "struct.Struct", other: Union["struct.Struct", float]
) -> "struct.Struct":
    """Returns self + other"""
    return _binary_op(self, other, getattr(np.ndarray, "__add__"))


@overload
def __radd__(self: "struct.Cube", other: float) -> "struct.Cube":
    ...


@overload
def __radd__(self: "struct.Map", other: float) -> "struct.Map":
    ...


@overload
def __radd__(self: "struct.Profile", other: float) -> "struct.Profile":
    ...


def __radd__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns other + self"""
    return self.__add__(other)


@overload
def __sub__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __sub__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __sub__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __sub__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __sub__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __sub__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self - other"""
    return _binary_op(self, other, getattr(np.ndarray, "__sub__"))


@overload
def __rsub__(self: "struct.Cube", other: float) -> "struct.Cube":
    ...


@overload
def __rsub__(self: "struct.Map", other: float) -> "struct.Map":
    ...


@overload
def __rsub__(self: "struct.Profile", other: float) -> "struct.Profile":
    ...


def __rsub__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns other - self"""
    return self.__sub__(other).__neg__()


@overload
def __mul__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __mul__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __mul__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __mul__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __mul__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __mul__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self * other"""
    return _binary_op(self, other, getattr(np.ndarray, "__mul__"))


@overload
def __rmul__(self: "struct.Cube", other: float) -> "struct.Cube":
    ...


@overload
def __rmul__(self: "struct.Map", other: float) -> "struct.Map":
    ...


@overload
def __rmul__(self: "struct.Profile", other: float) -> "struct.Profile":
    ...


def __rmul__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns other * self"""
    return self.__mul__(other)


@overload
def __truediv__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __truediv__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __truediv__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __truediv__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __truediv__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __truediv__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self / other"""
    return _binary_op(self, other, getattr(np.ndarray, "__truediv__"))


@overload
def __rtruediv__(self: "struct.Cube", other: float) -> "struct.Cube":
    ...


@overload
def __rtruediv__(self: "struct.Map", other: float) -> "struct.Map":
    ...


@overload
def __rtruediv__(self: "struct.Profile", other: float) -> "struct.Profile":
    ...


def __rtruediv__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns other / self"""
    return _binary_op(self, other, getattr(np.ndarray, "__rtruediv__"))


@overload
def __floordiv__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __floordiv__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __floordiv__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __floordiv__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __floordiv__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __floordiv__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self // other"""
    return _binary_op(self, other, getattr(np.ndarray, "__floordiv__"))


@overload
def __mod__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __mod__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __mod__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __mod__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __mod__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __mod__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self % other"""
    return _binary_op(self, other, getattr(np.ndarray, "__mod__"))


@overload
def __pow__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __pow__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __pow__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __pow__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __pow__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __pow__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self ** other"""
    return _binary_op(self, other, getattr(np.ndarray, "__pow__"))


@overload
def __rpow__(self: "struct.Cube", other: float) -> "struct.Cube":
    ...


@overload
def __rpow__(self: "struct.Map", other: float) -> "struct.Map":
    ...


@overload
def __rpow__(self: "struct.Profile", other: float) -> "struct.Profile":
    ...


def __rpow__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns other ** self"""
    return _binary_op(self, other, getattr(np.ndarray, "__rpow__"))


# Unary boolean operators


@overload
def __invert__(self: "struct.Cube") -> "struct.Cube":
    ...


@overload
def __invert__(self: "struct.Map") -> "struct.Map":
    ...


@overload
def __invert__(self: "struct.Profile") -> "struct.Profile":
    ...


def __invert__(self: "struct.Struct"):
    """Returns ~self"""
    if not self.is_logical():
        raise TypeError(
            "Self must be a logical structure i.e. to contain only 0 and 1 samples."
        )
    return type(self)(1 - self.data.astype("int"), self.header)


# Binary boolean operators


@overload
def __or__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __or__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __or__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __or__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __or__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __or__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self | other"""
    if not self.is_logical():
        raise TypeError(
            "Self must be a logical structure i.e. to contain only 0 and 1 samples."
        )
    if not self.is_logical():
        raise TypeError(
            "Other must be a logical structure i.e. to contain only 0 and 1 samples."
        )
    return _binary_op(self, other, getattr(np, "maximum"))


@overload
def __ror__(self: "struct.Cube", other: float) -> "struct.Cube":
    ...


@overload
def __ror__(self: "struct.Map", other: float) -> "struct.Map":
    ...


@overload
def __ror__(self: "struct.Profile", other: float) -> "struct.Profile":
    ...


def __ror__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns other | self"""
    return self.__or__(other)


@overload
def __and__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __and__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __and__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __and__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __and__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __and__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self & other"""
    if not self.is_logical():
        raise TypeError(
            "Self must be a logical structure i.e. to contain only 0 and 1 samples."
        )
    if not self.is_logical():
        raise TypeError(
            "Other must be a logical structure i.e. to contain only 0 and 1 samples."
        )
    return _binary_op(self, other, getattr(np, "minimum"))


@overload
def __rand__(self: "struct.Cube", other: float) -> "struct.Cube":
    ...


@overload
def __rand__(self: "struct.Map", other: float) -> "struct.Map":
    ...


@overload
def __rand__(self: "struct.Profile", other: float) -> "struct.Profile":
    ...


def __rand__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns other & self"""
    return self.__and__(other)


@overload
def __xor__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __xor__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __xor__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __xor__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __xor__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __xor__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self ^ other"""
    if not self.is_logical():
        raise TypeError(
            "Self must be a logical structure i.e. to contain only 0 and 1 samples."
        )
    if not self.is_logical():
        raise TypeError(
            "Other must be a logical structure i.e. to contain only 0 and 1 samples."
        )
    return _binary_op(self, other, getattr(np.ndarray, "__ne__"))


@overload
def __rxor__(self: "struct.Cube", other: float) -> "struct.Cube":
    ...


@overload
def __rxor__(self: "struct.Map", other: float) -> "struct.Map":
    ...


@overload
def __rxor__(self: "struct.Profile", other: float) -> "struct.Profile":
    ...


def __rxor__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns other ^ self"""
    return self.__xor__(other)


# Comparison operators


@overload
def __eq__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __eq__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __eq__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __eq__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __eq__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __eq__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self == other"""
    return _binary_op(self, other, getattr(np.ndarray, "__eq__"))


@overload
def __ne__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __ne__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __ne__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __ne__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __ne__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __ne__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self != other"""
    return _binary_op(self, other, getattr(np.ndarray, "__ne__"))


@overload
def __ge__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __ge__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __ge__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __ge__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __ge__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __ge__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self >= other"""
    return _binary_op(self, other, getattr(np.ndarray, "__ge__"))


@overload
def __gt__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __gt__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __gt__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __gt__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __gt__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __gt__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self > other"""
    return _binary_op(self, other, getattr(np.ndarray, "__gt__"))


@overload
def __le__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __le__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __le__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __le__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __le__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __le__(self: "struct.Struct", other) -> "struct.Struct":
    """Returns self <= other"""
    return _binary_op(self, other, getattr(np.ndarray, "__le__"))


@overload
def __lt__(
    self: "struct.Cube",
    other: Union["struct.Cube", "struct.Map", "struct.Profile", float],
) -> "struct.Cube":
    ...


@overload
def __lt__(self: "struct.Map", other: Union["struct.Map", float]) -> "struct.Map":
    ...


@overload
def __lt__(
    self: "struct.Map", other: Union["struct.Cube", "struct.Profile"]
) -> "struct.Cube":
    ...


@overload
def __lt__(
    self: "struct.Profile", other: Union["struct.Profile", float]
) -> "struct.Profile":
    ...


@overload
def __lt__(
    self: "struct.Profile", other: Union["struct.Cube", "struct.Map"]
) -> "struct.Cube":
    ...


def __lt__(
    self: "struct.Struct", other: Union["struct.Struct", float]
) -> "struct.Struct":
    """Returns self < other"""
    return _binary_op(self, other, getattr(np.ndarray, "__lt__"))


# Getitem operator


@overload
def __getitem__(
    self: "struct.Cube",
    index: Union[
        int, Tuple[int], np.ndarray, "struct.Cube", "struct.Map", "struct.Profile"
    ],
) -> np.ndarray:
    ...


@overload
def __getitem__(
    self: "struct.Map", index: Union[Tuple[int], np.ndarray, "struct.Map"]
) -> np.ndarray:
    ...


@overload
def __getitem__(
    self: "struct.Profile", index: Union[int, np.ndarray, "struct.Profile"]
) -> np.ndarray:
    ...


def __getitem__(
    self: "struct.Struct", index: Union[int, Tuple[int], np.ndarray, "struct.Struct"]
) -> np.ndarray:
    """Returns self[index]"""
    if isinstance(index, struct.Struct):
        if not index.is_logical():
            warn("index is not a logical structure but it is casted as it")
        index_ = index.data.astype(bool)

    if isinstance(self, struct.Cube):
        if isinstance(index, Integral):
            return self.data[index]  # 2D array
        if isinstance(index, tuple):
            return self.data[:, index[1], index[0]]  # 1D array
        if isinstance(index, np.ndarray):
            return self.data[
                index
            ]  # returned array shape depends on index shape and type
        if isinstance(index, struct.Cube):
            return self.data[index_]  # 1D array
        if isinstance(index, struct.Map):
            return self.data[:, index_]  # 2D array
        if isinstance(index, struct.Profile):
            return self.data[index_, :, :]  # 2D array
        raise ValueError(
            "index must be an instance of int, tuple, ndarray, Cube, Map or Profile to be a valid index of type Cube"
        )
    if isinstance(self, struct.Map):
        if isinstance(index, tuple):
            return self.data[index[1], index[0]]  # 0D array
        if isinstance(index, np.ndarray):
            return self.data[
                index
            ]  # returned array shape depends on index shape and type
        if isinstance(index, struct.Map):
            return self.data[index_]  # 1D array
        raise ValueError(
            "index must be an instance of tuple, ndarray or Map to be a valid index of type Map"
        )
    if isinstance(self, struct.Profile):
        if isinstance(index, Integral):
            return self.data[index]  # 0D array
        if isinstance(index, np.ndarray):
            return self.data[
                index
            ]  # returned array shape depends on index shape and type
        if isinstance(index, struct.Profile):
            return self.data[index_]  # 1D array
        raise ValueError(
            "index must be an instance of int, ndarray or Profile to be a valid index of type Profile"
        )



from math import ceil, pi, sqrt
from typing import Literal, Optional, Tuple, Union

import numpy as np

from .. import structures as struct

__all__ = ["kde"]


def kde(
    input: "struct.Struct",
    axis: Optional[Literal["spatial", "spectral"]] = None,
    h: Optional[Union[float, np.ndarray]] = None,
    t: Optional[np.ndarray] = None,
    t_bounds: Optional[Tuple[float]] = None,
    t_step: float = 0.05,
) -> Tuple[np.ndarray]:
    """If axis == 'spectral' then the kde is computed over the spectral axis i.e. pixel-wise.
    Use a gaussian kernel of parameter h

    Parameters
    ----------
    input : Cube | Map | Profile
        Input structure.
    axis : str, optional
        Axis over which to compute the PDF. Must be set only if input is an instance of Cube.
        If axis is None, the PDF is computed over all the flattened structure so the output
        `pdf` is a 1D-array
    h : float or ndarray, optional
        Kernel parameter. Must be a scalar if input is an instance of Map or Profile.
        If input is an instance of Cube, `h` must be an array of shape (nz,) if axis == 'spectral'
        or (ny, nx) if axis == 'spatial'.
    t : ndarray, optional
        Variable of the estimated PDF. Must be set if a specific range of value is needed.
    t_bounds : tuple of float, optional
        Bounds of `t`. Ignored if t is not None. Default : None.
    t_step : float, optional
        Step between two values of `t`. Ignored if t is not None. Default : 0.05.
        If `t_step` is not compatible with `t_step`, the right bound could be modified.

    Returns
    -------

    t : ndarray
        Variable of the estimated PDF.
    pdf : ndarray
        Estimated PDF.
    """
    if axis is not None and not isinstance(input, struct.Cube):
        raise ValueError("axis must only be set if input is an instance of Cube")
    elif axis is not None and isinstance(input, struct.Cube):
        axis = axis.lower()
        if axis == "spectral":
            data = np.moveaxis(input.data, 0, -1)
        elif axis == "spatial":
            data = input.data.reshape(input.nz, -1)
        else:
            raise ValueError(f"axis must only be 'spectral' or 'spatial', not {axis}")
    else:
        data = input.data.flatten()

    if h is None:
        h = 1.06 * np.nanstd(data, axis=-1) * data.shape[-1] ** (-1 / 5)
    else:
        h = float(h)

    if t_bounds is None:  # 5 sigma bound
        t_bounds = (data.min() - 5 * h.min(), data.max() + 5 * h.max())
    if t is None:
        n_t = ceil((t_bounds[1] - t_bounds[0]) / t_step)
        t = np.arange(n_t) * t_step + t_bounds[0]

    t_ = np.expand_dims(t, axis=tuple(range(data.ndim)))
    data_ = np.expand_dims(data, axis=-1)
    h_ = h if isinstance(h, float) else np.expand_dims(h, axis=(-1, -2))

    # pdf = norm.pdf(t_, loc = data_, scale = h_).mean(axis = -2)
    pdf = 1 / (sqrt(2 * pi) * h_) * np.exp(-((t_ - data_) ** 2) / (2 * h_**2))
    pdf = np.nanmean(pdf, axis=-2)

    return t, pdf
